"""
This module contains functions to generate random data
"""

import random
import string
import datetime


def random_string(length=6, chars=string.ascii_letters):
    """
    This function generates random strings with specified lengths.

    :param length: Length of the random string
    :param chars: Charset to choose chracters from
    :return: Randomly generated string

    Example:
        random_string(30, string.ascii_letters + string.digits)

    """
    return "".join(random.choice(chars) for _ in range(length))


def random_datetime(start, end):
    """
    This function will return a random datetime between the range
    of two datetime objects.

    :param start: Startdate
    :param end: Enddate
    :return: A random Datetime object

    Example:
        from_date = datetime.datetime(year=2001, month=1, day=1)
        to_date = datetime.datetime.now()
        random_datetime(from_date, to_date)

    """
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = random.randrange(int_delta)
    return start + datetime.timedelta(seconds=random_second)



import time
import unittest
from random import random

from mandarina.benchmark import Benchmark, start_timing, counter
from mandarina.benchmark import timer
from mandarina.benchmark import get_process_memory_usage


class TestBenchmark(unittest.TestCase):
    def test(self):
        pass

    def test_benchmark(self):
        execution_time = 0.1
        result = Benchmark.run(lambda: time.sleep(execution_time), 10, print_output=False)
        self.assertAlmostEqual(result[0], execution_time, 2)
        self.assertAlmostEqual(result[1], execution_time, 2)

    def test_timer(self):
        # Call to test if function runs without errors
        @timer
        def f():
            time.sleep(0.1)

    def test_get_process_memory_usage(self):
        # Call to test if function runs without errors
        get_process_memory_usage()
        get_process_memory_usage(readable=False)

    def test_start_timing(self):
        elapsed_time = start_timing()
        time.sleep(1)

        self.assertAlmostEqual(elapsed_time(), 1, 2)
        time.sleep(0.1)
        self.assertAlmostEqual(elapsed_time(), 1.1, 2)
        time.sleep(0.3)
        self.assertAlmostEqual(elapsed_time(), 1.4, 2)
        print(elapsed_time())

    def test_counter(self):
        @counter
        def f():
            pass

        f()
        self.assertEqual(f.calls, 1)
        for i in range(88):
            f()
        self.assertEqual(f.calls, 89)



from datetime import datetime


def frange(start, stop, increment):
    """
    Custom iterator to iterate in floating point steps.
    :return: Generator of steps between start and stop

    Example:
        for n in frange(0, 4, 0.5):
        print(n)
    """
    x = start
    while x < stop:
        yield x
        x += increment


def unixtime_to_isotime(unixtime, microseconds):
    """
    Converts a unix timestamp to a readable string timestamp
    :param unixtime:
    :param microseconds: If True microseconds will be visible
    :return:
    """
    return datetime.fromtimestamp(unixtime).strftime('%Y-%m-%dT%H:%M:%S' + microseconds * '.%f')

def apply_async(func, args, *, callback):
    """
    Async callback
    :param func:
    :param args:
    :param callback:
    :return:
    """
    # Compute the result
    result = func(*args)

    # Invoke the callback with the result
    callback(result)

def get_iso_timestamp():
    """
    Returns a string of an ISO 8601 timestamp without microseconds
    :return: ISO 8601 timestamp string
    """
    return datetime.datetime.now().replace(microsecond=0).isoformat()



"""
This module provides functionality for timing of methods.
"""

import time
import threading


def run_function_every_n_seconds(fn, fn_args, seconds):
    """
        Runs a function every n seconds.

        :param fn: The function to run
        :param fn_args: Arguments to pass to the function as list
        :param seconds: Timeinterval in seconds between function executions
        :return: None

        """
    start_time = time.time()
    while True:
        fn(fn_args)
        time.sleep(seconds - ((time.time() - start_time) % seconds))


def run_function_every_n_seconds_thread(fn, fn_args, seconds):
    thread = threading.Thread(name="run_function_every_n_seconds",
                              target=run_function_every_n_seconds, args=[fn, fn_args, seconds])

def run_function_after_n_seconds(fn, fn_args, seconds):
    """
    Runs a function after n seconds have passed
    :param fn:
    :param fn_args:
    :param seconds:
    :return:
    """
    start_time = time.time()
    time.sleep(seconds - ((time.time() - start_time) % seconds))
    fn(fn_args)


def run_function_after_n_seconds_thread(fn, fn_args, seconds):
    thread = threading.Thread(name="run_function_after_n_seconds",
                              target=run_function_after_n_seconds, args=[fn, fn_args, seconds])
    thread.start()



"""
This module provides functionality to work with images
"""


def overlay(background_image, foreground_image, x_offset, y_offset):
    """
    Overlays a png image on another image.

    :param background_image: OpenCv image to be overlaid with foreground image
    :param foreground_image: OpenCv image to overlay
    :param x_offset: Position of the overlay in x direction
    :param y_offset: Position of the overlay in y direction
    :return: Image with overlay

    Example:
        s_img = cv2.imread("foreground.png", -1)
        l_img = cv2.imread("background.png")
        img = overlay(l_img, s_img, 50, 50)
        cv2.imshow("Overlay", img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    """

    y1, y2 = y_offset, y_offset + foreground_image.shape[0]
    x1, x2 = x_offset, x_offset + foreground_image.shape[1]

    alpha_s = foreground_image[:, :, 3] / 255.0
    alpha_l = 1.0 - alpha_s

    for c in range(0, 3):
        background_image[y1:y2, x1:x2, c] = (alpha_s * foreground_image[:, :, c] +
                                             alpha_l * background_image[y1:y2, x1:x2, c])

    return background_image





from mandarina.file import *
import unittest


class FileTest(unittest.TestCase):
    def test(self):
        pass

    # def test_create_dir_if_doesnt_exist(self):
    #     dir_path = "./test"
    #     create_dir_if_doesnt_exist(dir_path)
    #     self.assertTrue(os.path.exists(dir_path))
    #     delete_dir(dir_path)

    def test_convert_size_bytes_to_human_readable_fomat(self):
        bytes = [1024 * i for i in range(5)]
        for size in bytes:
            result = convert_size_bytes_to_human_readable_format(size)



from mandarina.scheduling import *
import unittest


class SchedulingTest(unittest.TestCase):
    def test(self):
        pass

    def test_run_function_every_n_seconds_thread(self):
        def fn(*s):
            print(s)

        run_function_every_n_seconds_thread(fn, ["tick", "tock"], 1.0)


    def test_run_function_after_n_seconds(self):
        def fn(*s):
            print(s)

        run_function_after_n_seconds_thread(fn, ["tock", "tock"], 1.0)
        time.sleep(10)



from mandarina.csv import *

import unittest
import random
import datetime

from mandarina.image import *


class ImageTests(unittest.TestCase):
    def test(self):
        pass

    def test_overlay(self):
        pass




"""
This module contains code benchmarking tools.
"""

import time
import statistics
import psutil
import os

from functools import wraps
from mandarina.file import convert_size_bytes_to_human_readable_format


class Benchmark:
    """
    This class runs a benchmark on a function passed to the
    run method. It calls the function a specified number of times
    and calculates the mean and standard deviation across all runs.
    """

    @staticmethod
    def run(function, runs, print_output=True):
        """
        The run method prints the benchmark statistics when called
        with a function as parameter. If you set print_output to False,
        the function will only return the final value without printing
        anything to the console.

        :param function: The function to benchmark
        :return: The mean running time in seconds and its standard
                 deviation.

        Example
            Benchmark.run(lambda: time.sleep(1), 100)

        """
        timings = []
        if print_output:
            print("Runs Median Mean Stddev")
        for i in range(runs):
            startTime = time.time()
            function()
            seconds = time.time() - startTime
            timings.append(seconds)
            median = statistics.median(timings)
            mean = statistics.mean(timings)
            if print_output:
                if i < 10 or i % 10 == 9:
                    print(
                        "{0}\t{1:3.2f}\t{2:3.2f}\t{3:3.2f}".format(
                            1 + i,
                            median,
                            mean,
                            statistics.stdev(timings, mean) if i > 1 else 0,
                        )
                    )
        return (median, mean, statistics.stdev(timings, mean))


def timer(f):
    """
    Wraps a function in order to capture and print the
    execution time.

    Example
        @timer
        def f(x):
            print(x)

    """

    @wraps(f)
    def wrap(*args, **kwargs):
        start_time = time.time()
        result = f(*args, **kwargs)
        end_time = time.time()
        print(f"Function {f.__name__} with args {args, kwargs} took: {(end_time - start_time):.4f} seconds.")
        return result

    return wrap


def counter(func):
    """
    Wraps a function and counts the number of calls.

    Example:
        @counter
        def f(x):
            pass

        f(3)
        print(f.calls) # 1
        f(5)
        print(f.calls) # 2

    """

    @wraps(func)
    def helper(*args, **kwargs):
        helper.calls += 1
        return func(*args, **kwargs)

    helper.calls = 0
    return helper


def start_timing():
    """
    This closure function calculates the elapsed time in seconds since its initialization
    each time it is invoked.

    Example
        # Initialize
        elapsed_time = start_timing()

        time.sleep(1)

        # Get elapsed time
        et = elapsed_time()

    """
    CNT = [time.time()] * 2

    def calculate_elapsed_time():
        CNT[0] = CNT[1] - time.time()
        return abs(CNT[0])

    return calculate_elapsed_time


def get_process_memory_usage(readable=True):
    """
        Returns the memory usage of the current process the python
        script is running in.

        :param:  If readable is True the memory usage is returned
                 in a human readable format.
        :return: Memory usage in bytes or human readable format.

        Example
            get_process_memory_usage()

        """
    process = psutil.Process(os.getpid())
    if readable:
        return convert_size_bytes_to_human_readable_format(process.memory_full_info()[0])
    else:
        return process.memory_full_info()[0]



import unittest
from mandarina.randomdata import *

class RandomdataTests(unittest.TestCase):
    def test(self):
        pass

    def test_random_string(self):
        size = random.randint(0, 10 ** 6)
        charset = string.ascii_letters
        random_str = random_string(size)
        # Check if string has the correct size
        self.assertEqual(len(random_str), size)
        # Check if only allowed characters are in the random string
        self.assertTrue(set(random_str) <= set(charset))

    def test_random_datetime(self):
        from_date = datetime.datetime(year=2019, month=7, day=1)
        to_date = datetime.datetime.now()
        random_dt = random_datetime(from_date, to_date)
        self.assertGreaterEqual(random_dt, from_date)
        self.assertLessEqual(random_dt, to_date)



"""
This module provides functionality to work with files and folders.
"""
import math
import os
import fnmatch
import re
import hashlib

from pathlib import Path
from itertools import islice
from contextlib import redirect_stdout


def create_dir_if_doesnt_exist(folderpath):
    """
    Creates a folder if it doesn't exist.

    :param folderpath:
    :return: True if folder was created, else False

    """
    if not os.path.exists(folderpath):
        os.makedirs(folderpath)
        return True
    return False


def delete_file(filepath):
    """
    Deletes the specified file if the file exists.

    :param filepath: Path of the file to be deleted
    :return: True if the file was deleted, else False

    """
    if os.path.isfile(filepath):
        os.remove(filepath)
        return True
    else:
        print("Error: %s file not found" % filepath)
        return False


def delete_dir(dirpath):
    """
    Deletes the specified directory if it exists.

    :param dirpath: Path to the directory
    :return: True if the directory was deleted, else False

    """
    if os.path.isdir(dirpath):
        os.rmdir(dirpath)
        return True
    else:
        print("Error: %s directory not found" % dirpath)
        return False


def count_files_in_dir(dirpath):
    """
    Counts the number of files contained in the specified
    directory.

    :param dirpath: Path to the directory
    :return: Number of files

    """
    return len(
        [
            name
            for name in os.listdir(dirpath)
            if os.path.isfile(os.path.join(dirpath, name))
        ]
    )


def gen_find_files_wildcard(filepat, top):
    """
    Find all filenames in a directory tree that match a shell wildcard pattern.

    :param filepat: The pattern to find
    :param top: Top level directory to find in
    :return: Generator of found filepaths

    Example
        filenames = gen_find_files_wildcard("*.py", ".")

    """
    for path, dirlist, filelist in os.walk(top):
        for name in fnmatch.filter(filelist, filepat):
            yield os.path.join(path, name)


def gen_open_files(filepaths):
    """
    Open a sequence of filenames one at a time producing a file object.
    The file is closed immediately when proceeding to the next iteration.
    :param filenames: List of filepaths to open
    :return: Generator of opened files

    Example:
        files = gen_opener(filenames)
    """

    for filepath in filepaths:
        f = open(filepath, 'rt')
        yield f
        f.close()


def gen_concatenate(iterators):
    """
    Chain iterators together into a single sequence.
    :param iterators: Iterators to iterate over sequentially
    :return: Generator that iterates over all given iterators.

    Example:
        lines = gen_open_files(files)
    """
    for it in iterators:
        yield from it


def gen_write_file(out_filepath, generator_source):
    """
    Writes generator output to a file.
    :param out_filepath: File to write
    :param generator_source: Generator to consume from
    :return: None
    """
    with open(out_filepath, 'w') as f:
        for item in generator_source:
            f.write(str(item))


def lines_from_dir(filepat, dirname):
    """
    Creates a generator for all lines of all files matching a pattern
    in a given directory and its subdirectories
    :param filepat:
    :param dirname:
    :return:
    """
    filenames = gen_find_files_wildcard(filepat, dirname)
    files = gen_open_files(filenames)
    lines = gen_concatenate(files)
    return lines


def concatenate_files(top, pattern, out_file):
    """
    Writes files found by pattern to a single concatenated file
    :param top: Directory to search files in
    :param pattern: Pattern to find files by
    :return: True if file is written
    """
    lines = lines_from_dir(pattern, top)
    gen_write_file(out_file, lines)
    return os.path.isfile(out_file)



def gen_grep(pattern, lines):
    """
    Look for a regex pattern in a sequence of lines
    :param regex: The pattern to find
    :param lines: Lines to be matched
    :return: Lines with the matching pattern
    """
    pat = re.compile(pattern)
    for line in lines:
        if pat.search(line):
            yield line


def convert_size_bytes_to_human_readable_format(size_bytes):
    """
    Converts a size in bytes to a human readable format.

    :param size_bytes: The size in bytes
    :return: Bytes converted to readable format
    """
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


def md5(file_path):
    """
    Returns MD5 hash of file.

    :param file_path: The path to the file to be hashed
    :return: MD5 hash
    """
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



def tree(dir_path: Path, output_file_path, level: int=-1, limit_to_directories: bool=False,
         length_limit: int=1000):
    """
    Writes a graphical representation of the folder and file structure
    to a file, similar to the unix tree command.

    :return: None
    """
    with open(output_file_path, 'w') as f:
        with redirect_stdout(f):
            """Given a directory Path object print a visual tree structure"""
            dir_path = Path(dir_path) # accept string coerceable to Path
            files = 0
            directories = 0
            def inner(dir_path: Path, prefix: str='', level=-1):
                nonlocal files, directories
                if not level: 
                    return # 0, stop iterating
                if limit_to_directories:
                    contents = [d for d in dir_path.iterdir() if d.is_dir()]
                else: 
                    contents = sorted(list(dir_path.iterdir()))
                pointers = [tee] * (len(contents) - 1) + [last]
                for pointer, path in zip(pointers, contents):
                    if path.is_dir():
                        yield prefix + pointer + path.name
                        directories += 1
                        extension = branch if pointer == tee else space 
                        yield from inner(path, prefix=prefix+extension, level=level-1)
                    elif not limit_to_directories:
                        yield prefix + pointer + path.name + " " + md5(path)
                        files += 1
            print(dir_path.name)
            iterator = inner(dir_path, level=level)
            for line in islice(iterator, length_limit):
                print(line)
            if next(iterator, None):
                print(f'... length_limit, {length_limit}, reached, counted:')
            print(f'\n{directories} directories' + (f', {files} files' if files else ''))


from mandarina.csv import *

import unittest
import random
import datetime

from mandarina.file import delete_file


class CsvTests(unittest.TestCase):
    def test(self):
        pass

    def test_is_date_in_last_line(self):
        todays_date = str(datetime.datetime.now().date())
        with open("test.csv", "w") as filehandle:
            filehandle.write(str(datetime.datetime.now()) + "\n")
        self.assertTrue(is_date_in_last_line(todays_date, "test.csv"))
        delete_file("test.csv")

    # def test_delete_last_line(self):
    #     random_int = random.randint(0, 10)
    #     with open("test.csv", "w") as filehandle:
    #         for i in range(random_int):
    #             filehandle.write(str(datetime.datetime.now()) + "\n")
    #
    #     delete_last_line("test.csv")
    #     self.assertEqual(random_int - 1, count_lines("test.csv"))
    #     delete_file("test.csv")

    def test_create_headers(self):
        header_line = "This,is,a,test"
        create_headers("test_header.csv", header_line)
        with open("test_header.csv") as f:
            first_line = f.readline().strip()
        self.assertEqual(first_line, header_line)
        delete_file("test_header.csv")

    def test_count_lines(self):
        random_int = random.randint(1, 10 ** 6)
        with open("test.csv", "w") as filehandle:
            for i in range(random_int):
                filehandle.write(str(datetime.datetime.now()) + "\n")
        self.assertEqual(count_lines("test.csv"), random_int)
        delete_file("test.csv")



"""
This module provides functions to work with csv files.
"""

import subprocess
import codecs
import warnings
import os


def day_is_already_written(todays_date, filepath):
    warnings.warn(
        "WARNING: This function is deprecated, use the "
        "function is_date_in_last_line instead!"
    )
    return is_date_in_last_line(todays_date, filepath)


def is_date_in_last_line(todays_date, filepath):
    """
    Checks if the specified date is already written in the last line
    of a csv file. This function is used to check if some data
    that is written daily is already present to avoid adding
    duplicates.

    :param todays_date: Date as string like "2018-01-01"
    :param filepath: Filename to check
    :return: True if Date is present in the last line of the file

    Example
        is_date_in_last_line("2018-01-01", "data.csv")

    """
    line = subprocess.check_output(["tail", "-1", filepath])
    last_line = line.decode("utf-8")

    return todays_date in last_line


def is_file_empty(filepath):
    """
    Checks if a file has 0 lines.
    :param filepath: Path to the file
    :return: True if file has 0 lines, else False

    Example
        is_file_empty("data.csv")

    """
    return os.stat(filepath).st_size == 0


def delete_last_line(filepath):
    """
    Deletes the last line of the specified file.

    If the file has only one line, it is left.
    :param filepath: Filename to delete the last line from
    :return: None

    Example
        delete_last_line("data.csv")

    """
    with open(filepath, "r+", encoding="utf-8") as filehandle:
        # Move the pointer (similar to a cursor in a text editor) to the end of the file
        filehandle.seek(0, os.SEEK_END)
        # This code means the following code skips the very last character in the file -
        # i.e. in the case the last line is null we delete the last line
        # and the penultimate one
        pos = filehandle.tell() - 1
        # Read each character in the file one at a time from the penultimate
        # character going backwards, searching for a newline character
        # If we find a new line, exit the search
        while pos > 0 and filehandle.read(1) != "\n":
            pos -= 1
            filehandle.seek(pos, os.SEEK_SET)
        # So long as we're not at the start of the file, delete all the characters ahead
        # of this position
        if pos > 0:
            filehandle.seek(pos, os.SEEK_SET)
            filehandle.truncate()


def create_headers(filepath, header_row):
    """
    Creates a csv header row if the specified file is empty
    or doesn't exist.

    :param filepath: The filepath
    :param header_row: String for the header row
    :return True if file was empty and header is written

    Example
        create_headers("data.csv", "date,time,temperature,humidity")

    """
    with open(filepath, "a") as filehandle:
        # Check if file is empty and write row
        if os.stat(filepath).st_size == 0:
            filehandle.write(header_row + "\n")
            return True
        return False


def count_lines(filepath):
    """
    Counts the number of lines of a given file.

    :param file: The filepath
    :return: Number of lines contained in the file.

    Example
        count_lines("data.csv")

    """
    with codecs.open(filepath, encoding="utf-8", errors="ignore") as filehandle:
        if is_file_empty(filepath):
            return 0
        else:
            return len(filehandle.read().splitlines())
